# -*- coding: utf-8 -*-

import contextlib
import re

from django.db.utils import NotSupportedError
from pgtrigger import utils


class DatabaseSchemaEditorMixin:
    """
    A schema editor mixin that can subclass a DatabaseSchemaEditor and
    handle altering column types of triggers.

    Postgres does not allow altering column types of columns used in trigger
    conditions. Here we fix this with the following approach:

    1. Detect that a column type is being changed and set a flag so that we
       can alter behavior of the schema editor.
    2. In execute(), check for the special error message that's raised
       when trying to alter a column of a trigger. Temporarily drop triggers
       during the alter statement and reinstall them. Ensure this is all
       wrapped in a transaction
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.temporarily_dropped_triggers = set()
        self.is_altering_field_type = False

    @contextlib.contextmanager
    def alter_field_type(self):
        """
        Temporarily sets state so that execute() knows we are trying to alter a column type
        """
        self.is_altering_field_type = True
        try:
            yield
        finally:
            self.is_altering_field_type = False

    def _alter_field(
        self,
        model,
        old_field,
        new_field,
        old_type,
        new_type,
        old_db_params,
        new_db_params,
        strict=False,
    ):
        """
        Detects that a field type is being altered and sets the appropriate state
        """
        context = self.alter_field_type() if old_type != new_type else contextlib.nullcontext()

        with context:
            return super()._alter_field(
                model,
                old_field,
                new_field,
                old_type,
                new_type,
                old_db_params,
                new_db_params,
                strict=strict,
            )

    @contextlib.contextmanager
    def temporarily_drop_trigger(self, trigger, table):
        """
        Given a table and trigger, temporarily drop the trigger and recreate it
        after the context manager yields.
        """
        self.temporarily_dropped_triggers.add((trigger, table))

        try:
            with self.atomic, self.connection.cursor() as cursor:
                cursor.execute(
                    """
                    SELECT pg_get_triggerdef(oid) FROM pg_trigger
                        WHERE tgname = '%s AND tgrelid = %s::regclass;
                """,
                    (trigger, table),
                )
                trigger_create_sql = cursor.fetchone()[0]
                cursor.execute("DROP TRIGGER %s on %s;", (trigger, utils.quote(table)))

                yield

                cursor.execute(trigger_create_sql)
        finally:
            self.temporarily_dropped_triggers.remove((trigger, table))

    def execute(self, *args, **kwargs):
        """
        If we are altering a field type, catch the special error psycopg raises
        when a column on a trigger is altered. Temporarily drop and recreate
        triggers to ensure the alter operation is successful.
        """
        if self.is_altering_field_type:
            try:
                with self.atomic:
                    return super().execute(*args, **kwargs)
            except NotSupportedError as exc:
                error_message = exc.__cause__.pgerror
                table = self.get_table_name(exc.__cause__.cursor.query.decode())
                if table:
                    with self.atomic, self.connection.cursor() as cursor:
                        triggers = self.get_table_triggers(cursor, table)
                    for trigger in triggers:
                        if trigger in error_message:
                            if (table, trigger) not in self.temporarily_dropped_triggers:
                                with self.temporarily_drop_trigger(trigger, table):
                                    return self.execute(*args, **kwargs)
                raise  # pragma: no cover
        else:
            return super().execute(*args, **kwargs)

    def get_table_triggers(self, cursor, table_name, trigger_prefix="pgtrigger_%"):
        cursor.execute("""
            SELECT tgname
            FROM pg_trigger
            WHERE tgrelid = %s::regclass
                AND tgname LIKE %s;
        """, (table_name, trigger_prefix))
        trigger_names = [row[0] for row in cursor.fetchall()]
        return trigger_names

    def get_table_name(self, query: str):
        match = re.search(r'table "?(?P<table>\w+)"?', query, flags=re.IGNORECASE)
        if match:
            return match.groupdict()["table"]
